// anygetty/anygetty.c
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This is version $Name: V1-1-4 $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This file is taken 99% from mingetty.c:
//    Copyright (C) 1996 Florian La Roche  <laroche@redhat.com>
//    Copyright (C) 2002, 2003 Red Hat, Inc

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#define _GNU_SOURCE 1		       /* Needed to get setsid() */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <signal.h>
#include <fcntl.h>
#include <stdarg.h>
#include <ctype.h>
#include <utmp.h>
#include <getopt.h>
#include <sys/param.h>
#include <syslog.h>
#include <sys/utsname.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>


/* name of this program (argv[0]) */
static char *progname;
/* on which tty line are we sitting? (e.g. tty1) */
#define TTY_MAX_LEN 32
static char ttybuf[TTY_MAX_LEN];
static char* tty;
/* some information about this host */
static struct utsname uts;
/* the hostname */
static char hn[MAXHOSTNAMELEN + 1];
/* process and session ID of this program */
static pid_t pid, sid;
/* login program invoked */
static char *loginprog = "/bin/login";
/* Do not send a reset string to the terminal. */
static int noclear = 0;
/* Do not print a newline. */
static int nonewline = 0;
/* Do not print /etc/issue. */
static int noissue = 0;
/* Do not call vhangup() on the tty. */
static int nohangup = 0;
/* Do not print any hostname. */
static int nohostname = 0;
/* Print the whole string of gethostname() instead of just until the next "." */
static int longhostname = 0;
/* time to wait, seconds */
static int delay = 0;
/* chroot directory */
static char *ch_root = NULL;
/* working directory to change into */
static char *ch_dir = NULL;
/* 'nice' level of the program */
static int priority = 0;
/* automatic login with this user */
static char *autologin = NULL;
/* remote hostname */
static char* remote_hostname = NULL;

/* pid of child process */
pid_t child_pid;


/* error() - output error messages */
static void error (const char *fmt, ...)
{
	va_list va_alist;

	va_start (va_alist, fmt);
	openlog (progname, LOG_PID, LOG_AUTH);
	vsyslog (LOG_ERR, fmt, va_alist);
	/* no need, we exit anyway: closelog (); */
	va_end (va_alist);
	sleep (5);
	exit (EXIT_FAILURE);
}


static void make_id (char* id, int l)
{
  // Make an id based on the tty name following this rule (from man
  // utmp):
  // generate the ut_id by using the last two letters of
  // /dev/ttyp%c or by using p%d for /dev/pts/%d.
  // (Note that the latter part doesn't seem to be what Gnome Terminal
  // actually does; it uses ts/3 for pts/3 rather than the expected
  // p3)

  const char* x = tty;
  if (strncmp(x,"tty",3)==0) {
    x += 3;
  }
  if (strncmp(x,"pts/",4)==0) {
    x += 4;
    *(id++) = 'p';
    l--;
  }
  if (strlen(x)>l) {
    x += strlen(x) - l;
  }
  strncpy(id, x, l);
}


static void create_utmp (pid_t child_pid)
{
  // Create a new utmp entry; or, if a dead entry already exists for
  // this pty, reuse it.

  struct utmp ut;
  struct utmp *utp;
  time_t cur_time;

  char id[sizeof (ut.ut_id)];
  make_id(id,sizeof(id));

  setutent ();
  while ((utp = getutent ())) {
    if (utp->ut_type == DEAD_PROCESS
	&& strcmp(utp->ut_id,id)==0) {
      break;
    }
  }

  if (utp) {
    memcpy (&ut, utp, sizeof (ut));
  } else {
    memset (&ut, 0, sizeof (ut));
    strncpy (ut.ut_id, id, sizeof (ut.ut_id));
  }

  strncpy (ut.ut_user, "LOGIN", sizeof (ut.ut_user));
  strncpy (ut.ut_line, tty, sizeof (ut.ut_line));
  time (&cur_time);
  ut.ut_time = cur_time;
  ut.ut_type = LOGIN_PROCESS;
  ut.ut_pid = child_pid;
  ut.ut_session = sid;
  if (remote_hostname) {
    strncpy (ut.ut_host, remote_hostname, sizeof(ut.ut_host));
  } else {
    ut.ut_host[0]='\0';
  }

  pututline (&ut);
  endutent ();

  updwtmp (_PATH_WTMP, &ut);
}


static void clear_utmp ()
{
  // Clear our utmp entry, i.e. mark it as dead.

  struct utmp ut;
  struct utmp *utp = NULL;

  setutent ();
  while ((utp = getutent ())) {
    if (utp->ut_pid == child_pid) {
      break;
    }
  }

  if (!utp) {
    error("Couldn't find utmp record for pid %d",child_pid);
  }
	  
  memcpy (&ut, utp, sizeof (ut));
  ut.ut_type = DEAD_PROCESS;
  ut.ut_line[0]='\0';
  ut.ut_time = 0;
  ut.ut_user[0]='\0';
  ut.ut_host[0]='\0';
  
  pututline (&ut);
  endutent ();

  updwtmp (_PATH_WTMP, &ut);
}


void sighup_handler(int sig)
{
  clear_utmp();
  exit(0);
}


static void output_special_char (unsigned char c)
{
	switch (c) {
	case 's':
		printf ("%s", uts.sysname);
		break;
	case 'n':
		printf ("%s", uts.nodename);
		break;
	case 'r':
		printf ("%s", uts.release);
		break;
	case 'v':
		printf ("%s", uts.version);
		break;
	case 'm':
		printf ("%s", uts.machine);
		break;
	case 'o':
		printf ("%s", uts.domainname);
		break;
	case 'd':
	case 't':
		{
			time_t cur_time;
			struct tm *tm;
#if	0
			char buff[20];

			time (&cur_time);
			tm = localtime (&cur_time);
			strftime (buff, sizeof (buff),
				c == 'd'? "%a %b %d %Y" : "%X", tm);
			fputs (buff, stdout);
			break;
#else
			time (&cur_time);
			tm = localtime (&cur_time);
			if (c == 'd') /* ISO 8601 */
				printf ("%d-%02d-%02d", 1900 + tm->tm_year,
					tm->tm_mon + 1, tm->tm_mday);
			else
				printf ("%02d:%02d:%02d", tm->tm_hour,
					tm->tm_min, tm->tm_sec);
			break;
#endif
		}

	case 'l':
		printf ("%s", tty);
		break;
	case 'u':
	case 'U':
		{
			int users = 0;
			struct utmp *ut;
			setutent ();
			while ((ut = getutent ()))
				if (ut->ut_type == USER_PROCESS)
					users++;
			endutent ();
			printf ("%d", users);
			if (c == 'U')
				printf (" user%s", users == 1 ? "" : "s");
			break;
		}
	default:
		putchar (c);
	}
}

/* do_prompt - show login prompt, optionally preceded by /etc/issue contents */
static void do_prompt (int showlogin)
{
	FILE *fd;
	int c;

	if (nonewline == 0)
		putchar ('\n');
	if (noissue == 0 && (fd = fopen ("/etc/issue", "r"))) {
		while ((c = getc (fd)) != EOF) {
			if (c == '\\')
				output_special_char (getc (fd));
			else
				putchar (c);
		}
		fclose (fd);
	}
	if (nohostname == 0)
		printf ("%s ", hn);
	if (showlogin)
		printf ("login: ");
	fflush (stdout);
}

static char *get_logname (void)
{
	static char logname[40];
	char *bp;
	unsigned char c;

	tcflush (0, TCIFLUSH);		/* flush pending input */
	for (*logname = 0; *logname == 0;) {
		do_prompt (1);
		for (bp = logname;;) {
			if (read (0, &c, 1) < 1) {
				if (errno == EINTR || errno == EIO
					|| errno == ENOENT)
					exit (EXIT_SUCCESS);
				error ("%s: read: %s", tty, strerror (errno));
			}
			if (c == '\n' || c == '\r') {
				*bp = 0;
				break;
			} else if (!isprint (c))
				error ("%s: invalid character 0x%x in login"
					" name", tty, c);
			else if ((size_t)(bp - logname) >= sizeof (logname) - 1)
				error ("%s: too long login name", tty);
			else
				*bp++ = c;
		}
	}
	return logname;
}

static void usage (void)
{
	error ("usage: '%s [--noclear] [--nonewline] [--noissue] "
		"[--nohangup] [--nohostname] [--long-hostname] "
		"[--loginprog=/bin/login] [--nice=10] [--delay=10] "
		"[--chdir=/home] [--chroot=/chroot] [--autologin=user] "
	        "[--remotehost=hostname]"
		, progname);
}

static struct option const long_options[] = {
	{ "autologin", required_argument, NULL, 'a' },
	{ "chdir", required_argument, NULL, 'w' },
	{ "chroot", required_argument, NULL, 'r' },
	{ "delay", required_argument, NULL, 'd' },
	{ "noclear", no_argument, &noclear, 1 },
	{ "nonewline", no_argument, &nonewline, 1 },
	{ "noissue", no_argument, &noissue, 1 },
	{ "nohangup", no_argument, &nohangup, 1 },
	{ "no-hostname", no_argument, &nohostname, 1 }, /* compat option */
	{ "nohostname", no_argument, &nohostname, 1 },
	{ "loginprog", required_argument, NULL, 'l' },
	{ "long-hostname", no_argument, &longhostname, 1 },
	{ "nice", required_argument, NULL, 'n' },
	{ "remotehost", required_argument, NULL, 'h' },
	{ 0, 0, 0, 0 }
};

int main (int argc, char **argv)
{
	char *logname, *s;
	int c;

	progname = argv[0];
	if (!progname)
		progname = "mingetty";
	uname (&uts);
	gethostname (hn, MAXHOSTNAMELEN);
	hn[MAXHOSTNAMELEN] = '\0';
	pid = getpid ();
	sid = getsid (0);
#if	defined(s390) || defined(__s390__)
	putenv ("TERM=dumb");
#else
	putenv ("TERM=linux");
#endif

	while ((c = getopt_long (argc, argv, "a:d:l:n:w:r:", long_options,
		(int *) 0)) != EOF) {
		switch (c) {
		case 0:
			break;
		case 'a':
			autologin = optarg;
			break;
		case 'd':
			delay = atoi (optarg);
			break;
		case 'l':
			loginprog = optarg;
			break;
		case 'n':
			priority = atoi (optarg);
			break;
		case 'r':
			ch_root = optarg;
			break;
		case 'w':
			ch_dir = optarg;
			break;
		case 'h':
		        remote_hostname = optarg;
		        break;
		default:
			usage ();
		}
	}
	if (longhostname == 0 && (s = strchr (hn, '.')))
		*s = '\0';

	int rc = ttyname_r(0,ttybuf,TTY_MAX_LEN-1);
	if (rc) {
	  error("ttyname_r(0) failed: %s",strerror(rc));
	}
	tty=ttybuf;


	if (strncmp (tty, "/dev/", 5) == 0) /* ignore leading "/dev/" */
		tty += 5;

	if (delay)
		sleep (delay);

	if (autologin) {
		do_prompt (0);
		printf ("login: %s (automatic login)\n", autologin);
		logname = autologin;
	} else
		while ((logname = get_logname ()) == 0)
			/* do nothing */ ;

	if (ch_root)
		chroot (ch_root);
	if (ch_dir)
		chdir (ch_dir);
	if (priority)
		nice (priority);

	child_pid = fork();
	if (child_pid==-1) {
	  error("fork() failed: %s",strerror(errno));
	}
	if (child_pid==0) {
	  create_utmp(getpid());
	  execl (loginprog, loginprog, autologin? "-f" : "--", logname, NULL);
	  error ("%s: can't exec %s: %s", tty, loginprog, strerror (errno));
	}

	signal(SIGHUP,&sighup_handler);

	rc = waitpid(child_pid,NULL,0);
	if (rc==-1) {
	  error("waitpid() failed: %s",strerror(errno));
	}
	
	clear_utmp();
	exit (EXIT_SUCCESS);
}
