// apachemod/anyterm.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This is version $Name: V1-1-4 $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include <sstream>
#include <iomanip>

#include <unistd.h>
#include <signal.h>
#include <syslog.h>

#include <apr_general.h>  // for apr_generate_random_bytes

#include "anyterm.h"
#include "fifo.h"
#include "backend.h"
#include "html.h"
#include "editscript.h"
#include "wapterm.h"
#include "errors.h"
#include "shm_alloc.h"


using namespace std;


#define shm_size (200*1024)  /* Size of shared memory region */

#define ENABLE_STATUS 0  /* Enable ?status URI */


static Session::id_t random_id(void)
{
  // Danger danger danger!  The standard random() function does ODD
  // THINGS like returning the same number consecutively if used here,
  // presumably because of threads or multiple processes or some such.
  Session::id_t id;
  apr_generate_random_bytes((unsigned char*)&id, sizeof(id));
  return id;
}


Session::Session(int r, int c, int t, string cmd):
  id(random_id()),
  rows(r), cols(c), time_out(t), command(cmd),
  open_req(false), is_open(false), 
  screen(rows*cols),
  error(false),
  next(NULL), prev(NULL) {}


anyterm_globals* globals;


void spawn_rote_process(void)
{
  // Spawn the process containing the ROTE terminal emulations.

  globals->backend_pid=0;

  pid_t pgid = getpgid(0);

  int rc = fork();
  if (rc) {
    // Parent process

    // Backend sets its pid when it is ready
    while(!globals->backend_pid) {
      usleep(1000);
    }
    return;
  }

  // Child process

  // Put child process in same process group as parent.  This is
  // important, because Apache kills all its child processes by
  // signalling its process group.
  int pgrc = setpgid(0,pgid);
  if (pgrc==-1) {
    throw SysError("setpgid()");
  }

  try {
    backend_process(globals);
  }
  catch(Error& E) {
    syslog(LOG_ERR,"Anyterm backend process failed: %s",E.get_msg().c_str());
    exit(1);
  }
  catch(...) {
    syslog(LOG_ERR,"Anyterm backend process failed, reason unknown");
    exit(1);
  }
}


// The shared memory region:

key_t anyterm_shm_key;
int anyterm_shm_id;


void setup_backend_shmem(void)
{
  // If this is the first time an Anyterm request has been received,
  // create the shared memory segment and spawn the backend process.
  // If this is the first time an Anyterm request has been received by
  // this process, attach to the shared memory segment.
  // If Anyterm requests have previously been received by this process,
  // do nothing.

  if (anyterm_shm_id != -1) {
    // Not first in this process
    return;
  }

  bool exists = true;
  anyterm_shm_id = shmget(anyterm_shm_key, 0, 0600);
  if (anyterm_shm_id==-1) {
    if (errno==ENOENT) {
      exists = false;
      anyterm_shm_id = shmget(anyterm_shm_key, shm_size, IPC_CREAT|IPC_EXCL|0600);
    }
  }
  if (anyterm_shm_id==-1) {
    throw SysError("shmget()");
  } 

  void* addr = shmat(anyterm_shm_id, NULL, 0);
  if (reinterpret_cast<intptr_t>(addr)==-1) {
    throw SysError("shmat()");
  }


  // If the shared memory was created, we also need to initialise it
  // and start the backend process containing the ROTE terminal
  // emulations.
  if (!exists) {
    Shm::initialise_first(addr, shm_size);
    globals = new anyterm_globals();
    spawn_rote_process();    

  } else {
    Shm::initialise_subsq(addr);
    globals = reinterpret_cast<anyterm_globals*>(Shm::get_obj0());
  }
}



// Tell the backend that we have asked it to do something:

void signal_backend(void)
{
  kill(globals->backend_pid,SIGUSR1);
}


// Manage the session list

void anyterm_globals::add_session(Session* s)
{
  Writer l(globals->session_list_lock);
  s->prev=NULL;
  if (globals->sessions) {
    globals->sessions->prev=s;
  }
  s->next=globals->sessions;
  globals->sessions=s;
}


void anyterm_globals::remove_session(Session* s)
{
  Writer l(globals->session_list_lock);
  if (s->next) {
    s->next->prev=s->prev;
  }
  if (s->prev) {
    s->prev->next=s->next;
  } else {
    sessions=s->next;
  }
}


// Timeout unused sessions:

void Session::touch(void)
{
  last_access = time(NULL);
}


// Check if session has a backend error to report

void Session::report_any_backend_error(void)
{
  if (error) {
    error=false;
    throw Error(error_msg);
  }
}


// Handle Apache requests:

string int_to_string(int i)
{
  char b[32];
  snprintf(b,sizeof(b),"%d",i);
  return b;
}


Session* session_open(string command, int rows, int cols, int time_out)
{
  Session* ses = new Session(rows,cols,time_out,command);

  ses->touch();

  globals->add_session(ses);

  ses->open_req=true;  
  signal_backend();

  for (int c=0; c<10000; c++) {
    ses->report_any_backend_error();
    if (ses->is_open) {
      break;
    }
    usleep(1000);
  }

  if (!ses->is_open) {
    throw SysError("Session failed to open");
  }

  return ses;
}


static string anyterm_open(string command, int rows, int cols)
{
  Session* ses = session_open(command, rows, cols, ANYTERM_TIMEOUT);
  return "<open s=\"" + ses->id.str() + "\"/>";
}


string Session::close(void)
{
  open_req=false;
  signal_backend();

  for (int c=0; c<10000; c++) {
    report_any_backend_error();
    if (!is_open) {
      break;
    }
    usleep(1000);
  }

  if (is_open) {
    throw Error("Session failed to close");
  }

  globals->remove_session(this);
  delete this;

  return "<close/>";
}


static char decode_percent_escape(string s)
{
  char hexcode[3];
  hexcode[0]=s[1];
  hexcode[1]=s[2];
  hexcode[2]='\0';
  return strtol(hexcode,NULL,16);
}


string decode_uri_escapes(string s)
     // Decoded string is always <= length of input string
     // so *t should be allocated as the same length as s
     // including its terminating NULL
{
  string t;
  string::size_type i=0;
  while (i<s.length()) {
    switch (s[i]) {
    case '+': t+=' '; break;
    case '%': t+=decode_percent_escape(string(s,i,3)); i+=2; break;
    default: t+=s[i]; break;
    }
    i++;
  }
  return t;
}


string Session::send(string k)
{
  if (!is_open) {
    throw Error("not open");
  }

  if (!k.empty()) {
    string dk = decode_uri_escapes(k);
    unsigned int f = key_fifo.push(dk.data(),dk.length());
    signal_backend();
    if (f<dk.length()) {
      throw Error("characters lost");
    }
  }

  touch();

  return "<ip/>";
}



string escape_html(string s)
{
  string t;
  for(string::size_type i=0; i<s.length(); i++) {
    char c=s[i];
    switch(c) {
    case '<': t+="&lt;"; break;
    case '>': t+="&gt;"; break;
    case '&': t+="&amp;"; break;
    default:  t+=c; break;
    }
  }
  return t;
}


string Session::rcv(void)
{
  if (!is_open) {
    throw Error("not open");
  }

  for (int c=0; c<10000; c++) {
    touch();
    if (dirty || !is_open || error) {
      break;
    }
    usleep(1000);
  }

  report_any_backend_error();

  if (!is_open) {
    throw Error("session closed while waiting");
  }

  string html_screen;
  {
    Reader l(screen_lock);
    dirty=false;
    html_screen = htmlify_screen(this);
  }

  string editscript = make_editscript(old_html_screen,html_screen);
  old_html_screen=html_screen;

  return "<op>" + escape_html(editscript) + "</op>";
}


Session* get_session(Session::id_t id) 
{
  Reader l(globals->session_list_lock);
  for (Session* ses = globals->sessions;
       ses; ses = ses->next) {
    if (ses->id == id) {
      return ses;
    }
  }
  throw Error("No such session");
}


Session::id_t extract_ses(string args)
{
  string::size_type p = args.find("s=");
  if (p==args.npos) {
    return 0;
  }
  p+=2;
  string::size_type q = args.find("&",p);
  if (q==args.npos) {
    return args.substr(p);
  } else {
    return args.substr(p,q-p);
  }
}

uint64_t hexstr_to_uint64(string s)
{
  stringstream ss;
  ss << s;
  uint64_t i;
  ss >> setbase(16) >> i;
  return i;
}


string uint64_to_hexstr(uint64_t i)
{
  stringstream ss;
  ss << setbase(16) << i;
  return ss.str();
}



static int extract_num_arg(string args, string argname, int def)
{
  string::size_type p = args.find(argname);
  if (p==args.npos) return def;
  p+=argname.length();
  return strtol(args.substr(p).c_str(),NULL,10);
}


void report_session(Session* ses, ostream& s)
{
  // Reporting the random session ID is a security problem, as it
  // could be used by an attacker to interfere with someone else's
  // session.  For this reason, the following shows only the first
  // three hex digits of the ID.

  s << "Session ID: " << ses->id.str().substr(0,3) << "xxxxxxxxxxxxx\n"
    << "Dimensions: " << ses->rows << "x" << ses->cols << "\n"
    << "Command: " << static_cast<string>(ses->command) << "\n"
    << "Open? " << ses->is_open << "\n"
    << "Error? " << ses->error << "\n"
    << "Last error message: " << static_cast<string>(ses->error_msg) << "\n"
    << "\n";
}


#if ENABLE_STATUS
static string report_status(void)
{
  ostringstream s;
  s << "Backend process: " << globals->backend_pid << "\n\n"
    << "Current sessions:\n\n";
  {
    Reader l(globals->session_list_lock);
    for (Session* ses = globals->sessions;
	 ses; ses=ses->next) {
      report_session(ses,s);
    }
  }
  Shm::report_status(s);
  return s.str();
}

static response_t txt_resp(string s)
{
  return make_pair("text/plain",s);
}
#endif


static response_t xml_resp(string s)
{
  return make_pair("text/xml; charset=\"" CHARSET "\"", s);
}

response_t process_request(string args, string command)
{
  // Parse the arguments and call the appropriate function
  if (args.substr(0,3)=="wap") {
    return wapterm(args,command);
  }

  try {

    if (args.substr(0,4)=="open") {
      return xml_resp(anyterm_open(command,
				   extract_num_arg(args,"rows=",25),
				   extract_num_arg(args,"cols=",80)));

    } else if (args.substr(0,5)=="close") {
      return xml_resp(get_session(extract_ses(args))->close());
      
    } else if (args.substr(0,4)=="send") {
      string k = args.substr(args.find("k=")+2);
      return xml_resp(get_session(extract_ses(args))->send(k));

    } else if (args.substr(0,3)=="rcv") {
      return xml_resp(get_session(extract_ses(args))->rcv());
      
#if ENABLE_STATUS
    } else if (args.substr(0,6)=="status") {
      return txt_resp(report_status());
#endif

    } else {
      throw Error("invalid args '"+args+"'");
    }

  }
  catch (Error& E) {
    return xml_resp("<error>"+E.get_msg()+"</error>");
  }
}
