// apachemod/apache.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This is version $Name: V1-1-4 $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "apache.h"

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_main.h>
#include <http_protocol.h>
#include <http_request.h>
#include <util_script.h>
#include <http_connection.h>
#include <apr_strings.h>
#include <apr_tables.h>

#include <sys/types.h>
#include <sys/ipc.h>

#include "anyterm.h"
#include "errors.h"

using namespace std;


extern module AP_MODULE_DECLARE_DATA anyterm;

// Apache configuration commands:

typedef struct {
  char* command;
} anyterm_dir_config_t;

static void* anyterm_create_dir_cfg(apr_pool_t* pool, char* x)
{
  anyterm_dir_config_t* cfg = (anyterm_dir_config_t*)apr_palloc(pool, sizeof(anyterm_dir_config_t));
  cfg->command="/bin/bash --login";
  return cfg;
}


// Expand command string:
// %h -> remote hostname
// %% -> %

static string expand_command(string templ, request_rec* r)
{
  string::size_type p = templ.find('%');
  if (p==templ.npos || p==templ.length()-1) {
    return templ;
  }
  string v;
  switch(templ[p+1]) {
  case '%': v="%"; break;
  case 'h': v=ap_get_remote_host(r->connection,r->per_dir_config,REMOTE_NAME,NULL); break;
  case 'u': v=(r->user?r->user:""); break;
  default: v="?"; break;
  }

  return templ.substr(0,p) + v + expand_command(templ.substr(p+2),r);
}


// This is our entry point for each HTTP request:

static int anyterm_handler(request_rec* r)
{
  // If request does not have handler "anyterm", decline it so that
  // other handlers get the chance to look at it.
  if (string(r->handler)!="anyterm") {
    return DECLINED;
  }

  try {

    // Prevent caching.
    apr_table_set(r->headers_out,
		  "Cache-Control", "no-cache, no-store");
    apr_table_set(r->headers_out,
		  "Pragma", "no-cache");

    // Make sure our shared memory and backend process are set up
    setup_backend_shmem();
    
    // Find the URI's arguments (i.e. after the ?)
    string args;
    if (r->args) {
      args=r->args;
    } else {
      args="";
    }

    string command;
    if (args.substr(0,4)=="open" || args.substr(0,3)=="wap") {
      anyterm_dir_config_t* config = 
	(anyterm_dir_config_t*)ap_get_module_config(r->per_dir_config, &anyterm);
      command = expand_command(config->command,r);
    }
    
    response_t resp = process_request(args, command);
    
    ap_set_content_type(r,resp.first.c_str());
    if (resp.first.substr(0,8)=="text/xml") {
      ap_rputs("<?xml version=\"1.0\" encoding=\"" CHARSET "\"?>\n",r);
    }
    ap_rputs(resp.second.c_str(),r);
    
    return OK;
  }
  catch(Error& E) {
    ap_set_content_type(r,"text/plain");
    ap_rputs(E.get_msg().c_str(),r);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
}



// Apache initialisation:

static int anyterm_post_config(apr_pool_t *pconf, apr_pool_t *plog,
			       apr_pool_t *ptemp, server_rec *s)
{
  // This is called by Apache about once when the module is loaded.

  // Choose a key for the shared memory segment, which will be
  // available to all spawned processes.
  anyterm_shm_key = time(NULL);

  return OK;
}


static void anyterm_child_init(apr_pool_t *p, server_rec *s)
{
  // This is called by Apache each time a new request-handling process
  // is started.

  // Mark anyterm_shm_id so that setup_backend_shmem can know that it
  // is in a new process.
  anyterm_shm_id = -1;
}


static void anyterm_register_hooks(apr_pool_t *p)
{
  ap_hook_post_config(anyterm_post_config, NULL, NULL, APR_HOOK_MIDDLE); 
  ap_hook_child_init(anyterm_child_init, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler(anyterm_handler, NULL, NULL, APR_HOOK_MIDDLE);
}


static const command_rec anyterm_command_table[] = {
  AP_INIT_TAKE1("anyterm_command",
 		(const char*(*)())ap_set_string_slot,
 		(void*)APR_OFFSETOF(anyterm_dir_config_t,command),
 		OR_ALL,
 		"Command to run inside terminal"
 		),

  {NULL}
};

module AP_MODULE_DECLARE_DATA anyterm = {

  // Normally one uses the macro "STANDARD20_MODULE_STUFF" here, but
  // that uses __FILE__ to get the module name.  This file is called
  // "apache.cc" so that is not ideal.  Instead, write it out in full
  // and give "anyterm" as the module name.
  MODULE_MAGIC_NUMBER_MAJOR,
  MODULE_MAGIC_NUMBER_MINOR,
  -1,
  "anyterm",
  NULL,
  NULL,
  MODULE_MAGIC_COOKIE,
  NULL,
  
  anyterm_create_dir_cfg, /* per-directory config creator */
  NULL, /* dir config merger */
  NULL, /* server config creator */
  NULL, /* server config merger */
  anyterm_command_table,   /* command table */
  anyterm_register_hooks,  /* set up other request processing hooks */
};


