// apachemod/backend.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This is version $Name: V1-1-4 $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "backend.h"

#include <iconv.h>
#include <unistd.h>
#include <fcntl.h>

#include "anyterm.h"
#include "fifo.h"
#include "errors.h"


static iconv_t utf8_to_CHARSET;


struct anyterm_backend {
  // Private backend per-session state.
  RoteTerm* rt;
};


static bool rt_was_dirty(Session* ses)
{
  // Has the ROTE terminal's screen changed since this was last called?

  bool d = false;
  for (int r=0; r<ses->be->rt->rows; r++) {
    if (ses->be->rt->line_dirty[r]) {
      d=true;
      ses->be->rt->line_dirty[r]=false;
    }
  }
  if (ses->be->rt->curpos_dirty) {
    d=true;
    ses->be->rt->curpos_dirty=false;
  }
  return d;
}


static void scrape_screen(Session* ses)
{
  // Copy the ROTE terminal's screen into the shared memory.

  RoteCell* p = ses->screen.data();
  for (int r=0; r<ses->rows; r++) {
    for (int c=0; c<ses->cols; c++) {
      RoteCell cell = ses->be->rt->cells[r][c];
      *(p++) = cell;
    }
  }
  ses->cur_row = ses->be->rt->crow;
  ses->cur_col = ses->be->rt->ccol;
}


static void report_backend_error(Session* ses, Error& E)
{
  ses->error_msg=E.get_msg();
  ses->error=true;
}

static void backend_open (Session* ses)
{
  ses->be = (anyterm_backend*)malloc(sizeof(struct anyterm_backend));
  if (!ses->be) {
    throw SysError("malloc()");
  }

  ses->be->rt = rote_vt_create(ses->rows,ses->cols);
  if (!ses->be->rt) {
    throw SysError("rote_vt_create()");
  }

  int rc = rote_vt_forkpty(ses->be->rt, ses->command.c_str());
  if (rc==-1) {
    throw SysError("rote_vt_forkpty()");
  }

  // Get SIGIO when pty is readable or writable
  int pty_m_fd = rote_vt_get_pty_fd(ses->be->rt);

  rc = fcntl(pty_m_fd,F_SETOWN,getpid());
  if (rc==-1) {
    throw SysError("fcntl(pty,F_SETOWN)");
  }

  rc = fcntl(pty_m_fd,F_SETFL,O_ASYNC);
  if (rc==-1) {
    throw SysError("fcntl(pty,F_SETFL,O_ASYNC)");
  }

  scrape_screen(ses);

  ses->dirty=true;
  ses->error=false;
  ses->is_open=true;
}


static void backend_close (Session* ses)
{
  kill(ses->be->rt->childpid,SIGHUP);
  rote_vt_forsake_child(ses->be->rt);
  rote_vt_destroy(ses->be->rt);
  ses->open_req=false;
  ses->is_open=false;
  ses->dirty=true;
  free(ses->be);
  ses->be=NULL; // flush out any bugs!
}


static bool is_readable(int fd)
{
  fd_set ifs;
  FD_ZERO(&ifs);
  FD_SET(fd,&ifs);

  struct timeval tvzero;
  tvzero.tv_sec=0;
  tvzero.tv_usec=0;

  return select(fd+1,&ifs,NULL,NULL,&tvzero);
}



static volatile bool signaled;

static void sig_handler(int sig)
{
  signaled=true;
}


static void service_rote(anyterm_globals* globals)
{
  // This is the loop for the ROTE process.  For each open session, It
  // receives keystrokes from the fifo, sends them to ROTE and thence
  // to the pty, asks ROTE to read from the pty and update its screen
  // image, and updates the screen image in shared memory.  If this
  // has changed it sets the dirty flag in the shared memory.  Opening
  // new sessions and closing old ones also happens along the way.
  // Currently it polls with short waits; half of the code to use
  // pselect() is done (the request processes send sigusr1 when they
  // want something done) but we don't have a way to get the pty fd
  // yet.

  sigset_t mask, oldmask;
  // The following can all return errors, but only if the signal
  // number is invalid - which it isn't.
  sigemptyset(&mask);
  sigaddset(&mask,SIGUSR1);
  sigaddset(&mask,SIGIO);
  sigaddset(&mask,SIGALRM);

  while (1) {

    alarm(2);
    sigprocmask(SIG_BLOCK,&mask,&oldmask);  // errors as above
    while (!signaled) {
      sigsuspend(&oldmask);
    }
    signaled=false;

    Session* ses;
    {
      Reader l(globals->session_list_lock);
      ses = globals->sessions;
    }

    while (ses) {

      try {
	
	if (ses->open_req && !ses->is_open) {
	  // Open new session if requested
	  backend_open(ses);
	}
	
	if (!ses->open_req && ses->is_open) {
	  // Close session if requested
	  backend_close(ses);
	}

	if ((time(NULL) - ses->last_access) > (ses->time_out+EXPUNGE_TIMEOUT)) {
	  // Session is very old, throw it away
	  globals->remove_session(ses);
	  Session* tmp = ses;
	  ses=ses->next;  // Hmm, locking...
	  delete tmp;
	  continue;
	}
	
	if (ses->is_open) {
	  // For open sessions:
	  // Check if timed out
	  
	  if ((time(NULL) - ses->last_access) > ses->time_out) {
	    backend_close(ses);
	    throw Error("Session timed out");
	  }
	    
	  // Send keypresses
	  while (!ses->key_fifo.empty()) {
	    
	    // Get from FIFO:
	    char buf1[16];
	    size_t n = ses->key_fifo.pop(buf1,sizeof(buf1));
	    
	    // Convert from UTF8 to CHARSET:
	    char buf2[16];
	    size_t m = sizeof(buf2);
	    char* buf1ptr = buf1;
	    char* buf2ptr = buf2;
	    int rc = iconv(utf8_to_CHARSET, &buf1ptr, &n, &buf2ptr, &m);
	    if (rc==-1) {
	      throw SysError("iconv(utf8_to_" CHARSET ")");
	    }
	    
	    // Send to terminal:
	    rote_vt_write(ses->be->rt,buf2,sizeof(buf2)-m);
	  }
	}
      }
      catch (Error& E) {
	report_backend_error(ses,E);
      }

      {
	Reader l(globals->session_list_lock);
	ses = ses->next;
      }

    }

    // Slight delay here allows us to collect a chunk of screen
    // updates and send them in one go, probably making the
    // responsiveness better not worse.
    // Try adjusting this value if you like.  Units are microseconds.
    usleep(10000);

    {
      Reader l(globals->session_list_lock);
      ses = globals->sessions;
    }

    while (ses) {

      try {

	if (ses->is_open) {
	  // For open sessions:
	  
	  // If screen has changed, copy it into the shared memory.
	  rote_vt_update(ses->be->rt);
	  if (rt_was_dirty(ses)) {
	    Writer l(ses->screen_lock);
	    scrape_screen(ses);
	    ses->dirty = true;
	  }
	  
	  // It's possible that rote_vt_update has not read everything
	  // from the pty - it returns early so that other tasks can get
	  // their turn.  But we have to detect when this has happened
	  // since we don't get another SIGIO in this case.
	  if (is_readable(rote_vt_get_pty_fd(ses->be->rt))) {
	    signaled=true;
	  }
	}
      }
      catch (Error& E) {
	report_backend_error(ses,E);
      }

      {
	Reader l(globals->session_list_lock);
	ses = ses->next;
      }

    }

    sigprocmask(SIG_UNBLOCK,&mask,NULL);  // errors as above
  }
}


static void reap_child(int sig)
{
  // Handler function for SIGCHLD in backend process
  // Note that apparently you might get only one SIGCHLD for multiple
  // subprocesses if the terminate at about the same time.  So you
  // can't just have
  //    wait(NULL);

  pid_t rc;
  do {
    rc = waitpid(-1, NULL, WNOHANG);
  } while (rc>0);

  // Don't have to re-install signal handler on Linux
}



static void termination_handler(int sig)
{
  // Destroy the shared memory segment
  shmctl(anyterm_shm_id, IPC_RMID, NULL);
  exit(0);
}


void backend_process(anyterm_globals* globals)
{
  // It seems that Apache can put signal handling in an odd state with
  // various things ignored; this is not what our child processes
  // expect.  Sanitise signals.

  // These signal() calls can return errors, but not under any
  // realistic circumstances that I can think of.  Ignore them.

  for (int i=1; i<NSIG; i++) {
    signal(i,SIG_DFL);
  }
  sigset_t sigset;
  sigfillset(&sigset);
  sigprocmask(SIG_UNBLOCK,&sigset,NULL);


  // We don't want child processes to become zombies when they terminate.
  struct sigaction sa;
  sa.sa_handler=&reap_child;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags=SA_NOCLDSTOP;
  sigaction(SIGCHLD,&sa,NULL);  // should check return value

  
  // We'll get SIGUSR1 when a front-end process wants us to do something,
  // and a SIGIO when something happens on a pty.
  // SIGALRM is for timeouts.
  signaled=false;
  signal(SIGUSR1,&sig_handler);
  signal(SIGIO,&sig_handler);
  signal(SIGALRM,&sig_handler);


  // Tidy up on termination
  signal(SIGTERM,&termination_handler);
  signal(SIGINT, &termination_handler);
  signal(SIGHUP, &termination_handler);


  // We will have inherited all sorts of file descriptors from Apache,
  // e.g. log files, sockets etc.  This doesn't do much harm until
  // they are inherited by a spawned process where they could even be
  // a security problem.  So just close them all now.
  struct rlimit files_limit;
  int rc = getrlimit(RLIMIT_NOFILE,&files_limit);
  if (rc==-1) {
    throw SysError("getrlimit(RLIMIT_NOFILE)");
  }
  int max_fds = files_limit.rlim_cur;
  for (int fd=0; fd<max_fds; fd++) {
    close(fd);
  }


  // We'll be converting keyboard characters from UTF-8 to CHARSET:
  utf8_to_CHARSET = iconv_open(CHARSET,"UTF-8");
  if (utf8_to_CHARSET==(iconv_t)(-1)) {
    throw SysError("iconv_open(" CHARSET ",UTF-8)");
  }


  // Front-end process waits for us to tell it our pid:
  globals->backend_pid = getpid();


  // Start updating the terminals
  service_rote(globals);
}

