// apachemod/shm_alloc.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This is version $Name: V1-1-4 $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "shm_alloc.h"

#include "lock.h"

#include <limits.h>


// If CHECK_ALLOCATIONS is set to 1 the shared memory allocation will
// be checked for consistency before and after each allocation or
// deallocation.  This adds a significant overhead to these
// operations, though it does not increase the overall big-O
// complexity as allocation is already linear in the number of blocks.
// Any errors are logged to syslog.  You can also see the current
// allocations in the status page, where any inconsistencies will be
// flagged.  Please report any faults that you find, especially if you
// can reproduce them!

// Otherwise, set it to 0.


#define CHECK_ALLOCATIONS 1

#if CHECK_ALLOCATIONS==1
#include <string>
#include <sstream>
#include <syslog.h>
#endif


namespace Shm {

  // Basic block management:
  
  // The memory region being managed is divided into a series of blocks.
  // Each block is either in use or free.  Two free blocks are never
  // adjacent.  Each block starts with a header.  The header indicates
  // whether the block is in use or free, gives its length, and also the
  // length of the previous block.  The first block has prev_length=0.
  // At the end of the region a dummy header has length=0.
  
  // This is a naive algorithm, since allocation is O(n) in the number
  // of blocks (it scans all of the blocks to find the best one).  Real
  // algorithms manage separate lists for blocks of different sizes.

  struct block_header {
    bool in_use;
    size_t length;
    size_t prev_length;
  };
  
  // A per-region mutex prevents simultaneous access.  The region
  // header includes this mutex plus the header for the first block.

  struct region_header {
    ReadWriteLock region_lock;
    block_header first_block;
  };


#if CHECK_ALLOCATIONS==1
#define MEMCHECK(r,what,where) check_memory(r,what,where);
  void check_memory(region r, std::string what, void* where);
#else
#define MEMCHECK(r,what,where)
#endif

  region manage_region(void* start, size_t length)
  {
    // Newly-managed region is a single large block
    // Header at start
    region_header* hdr = new(start) region_header;
    hdr->first_block.in_use = false;
    hdr->first_block.length = length - sizeof(region_header) - sizeof(block_header);
    hdr->first_block.prev_length = 0;

    // Trailer at end
    block_header* trlr = (block_header*) ((char*)start + length - sizeof(block_header));
    trlr->length = 0;
    trlr->prev_length = hdr->first_block.length;

    return hdr;
  }
  
  
  region get_region(void* start)
  {
    return reinterpret_cast<region>(start);
  }
  
  
  static unsigned int round_up_align(unsigned int x, unsigned int r)
    // return x rounded up to the next multiple of r
  {
    return (x+r-1) & (~(r-1));
  }
  
  
  void* alloc(region r, size_t sz)
  {
    if (sz==0) {
      sz=1;
    }
    sz = round_up_align(sz,8);
    
    MEMCHECK(r, "alloc starts", 0);

    void* ret;
    {
      Writer l(r->region_lock);

      // Iterate over all of the blocks in the region, keeping track of
      // the best block so far.  Best means begin free and being large
      // enough, and having the smallest spare space.  Terminate early
      // if a perfect block, i.e. no spare space, is found.
      
      block_header* p = &(r->first_block);
      block_header* best = NULL;
      unsigned int best_spare = UINT_MAX;
      
      while (p->length!=0) {
	if (!p->in_use && p->length>=sz) {
	  unsigned int spare = p->length - sz;
	  if (spare < best_spare) {
	    best = p;
	    best_spare = spare;
	  }
	  if (spare == 0) {
	    break;
	  }
	}
	p = (block_header*) ((char*)p + sizeof(block_header) + p->length);
      }
      
      // Out of memory
      if (!best) throw std::bad_alloc();
      
      // Split block, storing new header for spare section, if there is
      // enough spare space to fit a header.
      if (best_spare>sizeof(block_header)) {
	best->length = sz;
	block_header* n = (block_header*) ((char*)best + sizeof(block_header) + sz);
	n -> in_use = false;
	n -> length = best_spare - sizeof(block_header);
	n -> prev_length = sz;
	block_header* m = (block_header*) ((char*)n + sizeof(block_header) + n->length);
	m -> prev_length = n->length;
      }
      
      best->in_use = true;

      ret = (char*)best + sizeof(block_header);
    }

    MEMCHECK(r, "alloc done", ret);
    return ret;
  }
  
  
  void dealloc(region r, void* b)
  {
    MEMCHECK(r, "dealloc starts", b);

    {
      Writer l(r->region_lock);

      // Get header of block being deallocated, and mark it as free.
      block_header* hdr = (block_header*) ((char*)b - sizeof(block_header));
      hdr->in_use = false;
      
      // Get header of following block.
      block_header* n = (block_header*) ((char*)b + hdr->length);
      
      // If following block is free and is not the end of the region,
      // merge it with the block being deallocated.
      if (n->length && !n->in_use) {
	hdr->length += n->length + sizeof(block_header);
      }
      
      // If block being freed is not the start of the region, get the
      // previous block.  If it is free, merge this block with it.
      if (hdr->prev_length) {
	block_header* p = (block_header*) ((char*)hdr - hdr->prev_length - sizeof(block_header));
	if (!p->in_use) {
	  p->length += hdr->length + sizeof(block_header);
	  hdr=p;
	}
      }
      
      // Make the prev_length of the next block match this
      // possibly-expanded blocks's length.
      block_header* m = (block_header*) ((char*)hdr + sizeof(block_header) + hdr->length);
      m->prev_length = hdr->length;
    }
      
    MEMCHECK(r, "dealloc done", b);
  }
  
  
  void* get_block0(region r)
  {
    return (void*) ((char*)r + sizeof(region_header));
  }


  // C++ interface

  region the_region;

  void initialise_first(void* start, size_t length)
  {
    the_region = manage_region(start,length);
  }

  void initialise_subsq(void* start)
  {
    the_region = get_region(start);
  }

  void* get_obj0(void) {
    return get_block0(the_region);
  }


  void report_status(std::ostream& s, region r)
  {
    Reader l(r->region_lock);

    s << "Shared memory:\n";

    unsigned int nfree=0;
    unsigned int nused=0;
    size_t szfree=0;
    size_t szused=0;
    size_t maxfree=0;
    block_header* hdr = &(r->first_block);
    unsigned int prev_length=0;
    while (1) {
      if (hdr->prev_length!=prev_length) {
	s << "  *** prev_length mismatch, expected " << prev_length
	  << ", got " << hdr->prev_length << " ***\n";
      }
      prev_length=hdr->length;
      s << (void*)((char*)hdr+sizeof(block_header)) << ": " << hdr->length << " bytes ";
      if (!hdr->length) {
	s << "Trailer\n";
	break;
      }
      if (hdr->in_use) {
	s << "in use\n";
	nused++;
	szused+=hdr->length;
      } else {
	s << "free\n";
	nfree++;
	szfree+=hdr->length;
	maxfree=std::max(maxfree,hdr->length);
      }

      hdr = (block_header*)((char*)hdr + sizeof(block_header) + hdr->length);
    }
    s << "\n"
      << "Free space: " << szfree << " bytes in " << nfree << " blocks\n"
      << "  largest block " << maxfree << " bytes\n"
      << "Used space: " << szused << " bytes in " << nused << " blocks\n"
      << "\n";
  } 


#if CHECK_ALLOCATIONS==1
  void check_memory(region r, std::string what, void* where)
  {
    {
      Reader l(r->region_lock);

      block_header* hdr = &(r->first_block);
      unsigned int prev_length=0;
      while (1) {
	if (hdr->prev_length!=prev_length) {
	  break;
	}
	if (!hdr->length) {
	  return;
	}
	prev_length=hdr->length;
	hdr = (block_header*)((char*)hdr + sizeof(block_header) + hdr->length);
      }
    }
    
    std::ostringstream s;
    report_status(s,r);
    syslog(LOG_ERR,"Anyterm shared memory allocation error detected "
	   "while %s at %p:\n%s",
	   what.c_str(), where, s.str().c_str());
  }
#endif

};
