// apachemod/shm_alloc.h
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This is version $Name: V1-1-4 $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#ifndef shm_alloc_h
#define shm_alloc_h


#include <stddef.h>
#include <stdint.h>

#include <new>
#include <iterator>
#include <iostream>


namespace Shm {

  // Memory allocation functions for the shared memory region.

  // C interface.

  // The region must first be initialised by calling manage_region
  // with its start address and length.  This returns an opaque
  // region handle which can be passed to alloc and dealloc to
  // allocate and deallocate blocks within the region.
  
  // The same region can have different addresses in different
  // processes.  This implementation copes with that, since it does not
  // store any pointers internally; it just records offsets.  However
  // the pointers returned by alloc are safe to use only in the
  // process where they were allocated.
  
  // The region handle is itself a pointer, so it is NOT safe to
  // share it between processes.  Instead each process should get its
  // own handle.  The first process can use the handle returned by
  // manage_region, and other processes can call get_region()
  // passing the start address of the region in their address space.
  
  // To solve bootstrapping problems the address of the first block
  // allocated can be found using get_block0().
  
  struct region_header;
  typedef region_header* region;
  region manage_region(void* start, size_t length);
  region get_region(void* start);
  void* alloc(region r, size_t sz);
  void dealloc(region r, void* p);
  void* get_block0(region r);

  // C++ interface to the above.

  // The above functions can cope with multiple regions, and the
  // region handle is passed as a parameter.  However the C++ alocator
  // is best kept stateless, so we designate a single global shared
  // region.

  extern region the_region;

  // Initialisation functions for the first and subsequent processes:

  void initialise_first(void* start, size_t length);
  void initialise_subsq(void* start);

  // The address of the first-allocated block can be found to solve
  // bootstrapping problems.

  void* get_obj0(void);

  // Rather than store pointers to things in the shared memory region
  // we store their offsets relative to the start of the region.  This
  // is necessary because the location of the region in the memory map
  // can vary from one process to another.  These offsets are
  // implicitly converted to and from pointers as necessary.
  
  // Care is needed with NULL values; without special consideration,
  // converting a NULL pointer to an offset and then back to a pointer
  // in another process would give a non-NULL pointer in the other
  // process.  Instead, NULL pointers are converted to 0 offsets.
  // This is safe only because there is always a header at the start
  // of the shared region so no user object is ever allocated there.

  template <typename T>
    class offset {
    private:

    intptr_t o;

    T* getptr(void) const {
      return o ? (reinterpret_cast<T*>(o + reinterpret_cast<intptr_t>(the_region)))
	: NULL;
    }
      
    public:

    typedef T                          value_type;
    typedef size_t                     size_type;
    typedef ptrdiff_t                  difference_type;
    typedef value_type*                pointer;
    typedef const value_type*          const_pointer;
    typedef value_type&                reference;
    typedef const value_type&          const_reference;
    typedef std::random_access_iterator_tag iterator_category;

    offset(): o(0) {}

    offset(const offset<T>& ofs): o(ofs.o) {}

    offset(const T* p):
      o ( p ? (reinterpret_cast<intptr_t>(p) - reinterpret_cast<intptr_t>(the_region))
	    : 0 ) {}

    operator const void* (void) const { return getptr(); }

    operator void* (void) const { return getptr(); }

    operator const_pointer (void) const { return getptr(); }

    operator pointer (void) { return getptr(); }

    //    const_reference operator*(void) const { return * getptr(); }

    reference operator*(void) const { return * getptr(); }

    //    const_pointer operator->(void) const { return getptr(); }

    pointer operator->(void) const { return getptr(); }

    //operator bool (void) { return o; }

    bool operator== (const offset<T> &rhs) const {
      return o == rhs.o;
    }

    bool operator!= (const offset<T> &rhs) const {
      return !operator== (rhs);
    }

    bool operator< (const offset<T> &rhs) const {
      return o < rhs.o;
    }

    bool operator<= (const offset<T> &rhs) const {
      return !(*this > rhs);
    }

    bool operator> (const offset<T> &rhs) const {
      return rhs < *this;
    }

    bool operator>= (const offset<T> &rhs) const {
      return !(*this < rhs);
    }

    offset<T>& operator+= (difference_type i) {
      o += sizeof(T)*i;
      return *this;
    }

    offset<T>& operator-= (difference_type i) {
      return operator+=(-i);
    }

    offset<T>& operator++ () {
      return operator+=(1);
    }

    offset<T> operator++ (int) {
      offset<T> tmp = *this;
      operator++();
      return tmp;
    }

    offset<T>& operator-- () {
      return operator-=(1);
    }

    offset<T> operator-- (int) {
      offset<T> tmp = *this;
      operator--();
      return tmp;
    }

    offset<T> operator+ (difference_type i) const {
      return offset<T> (*this) += i;
    }

    offset<T> operator- (difference_type i) const {
      return offset<T> (*this) -= i;
    }

    ptrdiff_t operator- (const offset<T> &rhs) const {
      const_pointer p1(*this);
      const_pointer p2(rhs);
      return p1 - p2;
    }

  };


  // Shm::allocator is an allocator that can be used with standard
  // library things (e.g. containers) that need to allocate memory; it
  // invokes the shared memory allocation functions above, and uses
  // offsets instead of pointers.

  template <typename T>
    class allocator {
    public:
    typedef size_t          size_type;
    typedef ptrdiff_t       difference_type;
    typedef offset<T>       pointer;
    typedef const offset<T> const_pointer;
    typedef T&              reference;
    typedef const T&        const_reference;
    typedef T               value_type;
    
    template <typename T1> struct rebind {
      typedef allocator<T1> other;
    };
    
    allocator() throw() {}
    allocator(const allocator&) throw() {}
    template <typename T1> allocator(const allocator<T1>&) throw() {}
    ~allocator() throw() {}
    pointer address(reference x) const { return &x; }
    const_pointer address(const_reference x) const { return &x; }
    
    pointer allocate(size_type n, const void* hint = 0) {
      return static_cast<T*>(alloc(the_region, n*sizeof(T)));
    }
    
    void deallocate(pointer p, size_type n) {
      dealloc(the_region, p);
    }
      
    size_type max_size() const throw() { return size_t(-1) / sizeof(T); }
      
    void construct(pointer p, const T& val) { new(p) T(val); }
    void destroy(pointer p) { p->~T(); }
      
    template <typename T1>
      bool operator==(const allocator<T1>&) { return true; }
    
    template <typename T1>
      bool operator!=(const allocator<T1>&) { return false; }
  };
    
#if 0
  // Something I was copying defined this.  It doesn't compile.  Note
  // sure what the deal is.

  template<>
    class allocator<void> {
    typedef size_t             size_type;
    typedef ptrdiff_t          difference_type;
    typedef offset<void>       pointer;
    typedef const offset<void> const_pointer;
    typedef void               value_type;
    
    template <class T1> struct rebind {
      typedef allocator<T1> other;
    };
  };
#endif


  // Objects that will be stored in the shared memory region can be
  // instances of subclasses of Shm::Obj.  This class defines new and
  // delete operators that use the shared memory allocation functions.
  // Note: the in-memory representation of objects with virtual
  // methods includes additional pointers (for the "vtable").  These
  // will fail to work if stored in shared memory.

  class Obj {
  public:
    
    static void* operator new (size_t sz) {
      return allocator<Obj>().allocate(sz);
    }
    static void* operator new[] (size_t sz) {
      return allocator<Obj>().allocate(sz);
    }
    static void* operator new(size_t size, void* loc) {
      return ::operator new(size,loc);
    }
    static void* operator new[](size_t size, void* loc) {
      return ::operator new[](size,loc);
    }

    static void operator delete (void* p, size_t sz) {
      allocator<Obj>().deallocate(static_cast<Obj*>(p), sz);
    }
    static void operator delete[] (void* p, size_t sz) {
      allocator<Obj>().deallocate(static_cast<Obj*>(p), sz);
    }

    Obj() {}
    Obj(const Obj& o) {}

  };

  // Generate a debug / status report
  void report_status(std::ostream& s, region r = the_region);

};



#endif
